#!/usr/bin/python

# sawpt.py: Perturbation theory calculation for 4D SAW field theory
# (requires Python 2.6 or higher)

# Roland Bauerschmidt <roland@bauerschmidt.ca>
# David C. Brydges <db5d@math.ubc.ca>
# Gordon Slade <slade@math.ubc.ca>

# Changelog:
# - Correction of sign in [rg-loc, (1.52)]
# - Correction of Loc(phi(nabla phi))

from wick import *


# coupling constants
half   = Fraction(1,2)
g      = Coupling('g')
nu     = Coupling('\\nu')
nuplus = Coupling('\\nu_+')
y      = Coupling('y')
z      = Coupling('z')

# a complex Boson and Fermion field (\phi, \bar\phi, \psi, \psi\bar)
phix   = Field('\\phi',               'x', fermion=0, dimension=1)
phiy   = Field('\\phi',               'y', fermion=0, dimension=1)
phibx  = Field('\\bar\\phi',          'x', fermion=0, dimension=1)
phiby  = Field('\\bar\\phi',          'y', fermion=0, dimension=1)
psix   = Field('\\psi',               'x', fermion=1, dimension=1)
psiy   = Field('\\psi',               'y', fermion=1, dimension=1)
psibx  = Field('\\bar\\psi',          'x', fermion=1, dimension=1)
psiby  = Field('\\bar\\psi',          'y', fermion=1, dimension=1)

# Laplace operator applied to the Fields
lphix  = Field('-\\Delta \\phi',      'x', fermion=0, dimension=3)
lphiy  = Field('-\\Delta \\phi',      'y', fermion=0, dimension=3)
lphibx = Field('-\\Delta \\bar\\phi', 'x', fermion=0, dimension=3)
lphiby = Field('-\\Delta \\bar\\phi', 'y', fermion=0, dimension=3)
lpsix  = Field('-\\Delta \\psi',      'x', fermion=1, dimension=3)
lpsiy  = Field('-\\Delta \\psi',      'y', fermion=1, dimension=3)
lpsibx = Field('-\\Delta \\bar\\psi', 'x', fermion=1, dimension=3)
lpsiby = Field('-\\Delta \\bar\\psi', 'y', fermion=1, dimension=3)

# Nabla operator applied to the Fields
nphix  = Field('\\nabla \\phi',       'x', fermion=0, dimension=2)
nphiy  = Field('\\nabla \\phi',       'y', fermion=0, dimension=2)
nphibx = Field('\\nabla \\bar\\phi',  'x', fermion=0, dimension=2)
nphiby = Field('\\nabla \\bar\\phi',  'y', fermion=0, dimension=2)
npsix  = Field('\\nabla \\psi',       'x', fermion=1, dimension=2)
npsiy  = Field('\\nabla \\psi',       'y', fermion=1, dimension=2)
npsibx = Field('\\nabla \\bar\\psi',  'x', fermion=1, dimension=2)
npsiby = Field('\\nabla \\bar\\psi',  'y', fermion=1, dimension=2)


# quadratic and quartic supersymmetric combinations
taux   = phibx*phix + psibx*psix
tauy   = phiby*phiy + psiby*psiy
tau2x  = taux*taux
tau2y  = tauy*tauy
taulx  = (phibx*lphix + lphibx*phix + psibx*lpsix + lpsibx*psix)*half
tauly  = (phiby*lphiy + lphiby*phiy + psiby*lpsiy + lpsiby*psiy)*half
taunx  = nphibx*nphix + npsibx*npsix
tauny  = nphiby*nphiy + npsiby*npsiy


# This function depends on the fields and a string that labels
# the covariance. Subsequently, new functions, C, W, and Wplus, are
# defined that are the same as __C, but have the string that label
# the covariance set to appropriate values
def __C(f1, f2, C='C'):
	l = [
		( phix,   phibx, C+'(0)'), 
		( phiy,   phiby, C+'(0)'), 
		( phiy,   phibx, C+'(x-y)'), 
		( phix,   phiby, C+'(x-y)'), 
		( psix,   psibx, C+'(0)'), 
		( psiy,   psiby, C+'(0)'), 
		( psiy,   psibx, C+'(x-y)'), 
		( psix,   psiby, C+'(x-y)'), 
		(lphix,   phibx, '-\\Delta '+C+'(0)'),
		(lphiy,   phiby, '-\\Delta '+C+'(0)'),
		(lphix,   phiby, '-\\Delta '+C+'(x-y)'),
		(lphiy,   phibx, '-\\Delta '+C+'(x-y)'),
		( phix,  lphibx, '-\\Delta '+C+'(0)'),
		( phiy,  lphiby, '-\\Delta '+C+'(0)'),
		( phix,  lphiby, '-\\Delta '+C+'(x-y)'),
		( phiy,  lphibx, '-\\Delta '+C+'(x-y)'),
		(lpsix,   psibx, '-\\Delta '+C+'(0)'),
		(lpsiy,   psiby, '-\\Delta '+C+'(0)'),
		(lpsix,   psiby, '-\\Delta '+C+'(x-y)'),
		(lpsiy,   psibx, '-\\Delta '+C+'(x-y)'),
		( psix,  lpsibx, '-\\Delta '+C+'(0)'),
		( psiy,  lpsiby, '-\\Delta '+C+'(0)'),
		( psix,  lpsiby, '-\\Delta '+C+'(x-y)'),
		( psiy,  lpsibx, '-\\Delta '+C+'(x-y)'),
		(nphix,  nphibx, '\\nabla\\nabla^* '+C+'(0)'),
		(nphix,  nphiby, '\\nabla\\nabla^* '+C+'(x-y)'),
		(nphiy,  nphibx, '\\nabla\\nabla^* '+C+'(x-y)'),
		( phix,  nphiby, '\\nabla^* '+C+'(x-y)'),
		( phiy,  nphibx, '\\nabla '+C+'(x-y)'),
		(nphix,   phiby, '\\nabla '+C+'(x-y)'),
		(nphiy,   phibx, '\\nabla^* '+C+'(x-y)'),
		(npsix,  npsibx, '\\nabla\\nabla^* '+C+'(0)'),
		(npsix,  npsiby, '\\nabla\\nabla^* '+C+'(x-y)'),
		(npsiy,  npsibx, '\\nabla\\nabla^* '+C+'(x-y)'),
		( psix,  npsiby, '\\nabla^* '+C+'(x-y)'),
		( psiy,  npsibx, '\\nabla '+C+'(x-y)'),
		(npsix,   psiby, '\\nabla '+C+'(x-y)'),
		(npsiy,   psibx, '\\nabla^* '+C+'(x-y)'),
	]

	for rf1, rf2, rC in l:
		if str(f1) == str(rf1) and str(f2) == str(rf2):
			# this may not be needed anymore
			if rC[0] == '-':
				rC = rC[1:]
				sign = -1
			else:
				sign = 1

			return Coupling(rC) * sign



def __old__C(f1, f2, C='C'):
	l = [
		( phix.name,   phibx.name, C), 
		( psix.name,   psibx.name, C),
		(lphix.name,   phibx.name, '-\\Delta '+C),
		( phix.name,  lphibx.name, '-\\Delta '+C),
		(lpsix.name,   psibx.name, '-\\Delta '+C),
		( psix.name,  lpsibx.name, '-\\Delta '+C),
		(nphix.name,  nphibx.name, '\\nabla\\nabla^* '+C),
		( phix.name,  nphibx.name, '\\nabla^* '+C),
		(nphix.name,   phibx.name, '\\nabla '+C),
		(npsix.name,  npsibx.name, '\\nabla\\nabla^* '+C),
		( psix.name,  npsibx.name, '\\nabla^* '+C),
		(npsix.name,   psibx.name, '\\nabla '+C),
	]

	for rf1, rf2, rC in l:
		if f1.name == rf1 and f2.name == rf2:
			x = f1.pos
			y = f2.pos

			if x == y:
				xy = '0'
			elif x < y:
				xy = x+'-'+y
			else:
				xy = y+'-'+x

			# allow to have covariances with negative sign;
			# this may not be needed anymore
			if rC[0] == '-':
				rC = rC[1:]
				sign = -1
			else:
				sign = 1

			return Coupling(rC+'('+xy+')') * sign

# definition of covariances C, W, and W_+
C     = (lambda f1, f2: __C(f1,f2,'C'))
W     = (lambda f1, f2: __C(f1,f2,'W'))
Wplus = (lambda f1, f2: __C(f1,f2,'W_+'))

# definition of F as in rg-pt
def F(C, U, V):
	UE = U.E(C, -1)
	VE = V.E(C, -1)
	return (UE*VE).E(C, 1) - U*V

# definition of Loc as in rg-loc
def Loc(i, pos, dim=4):

	# i is a monomial: use definition from rg-loc
	if isinstance(i, Monomial):

		if i.dimension() > dim:
			return Monomial(0)

		fields = []
		for f in i.fields:
			fields.append(Field(f.name, pos=pos, fermion=f.fermion, dimension=f.dimension))
		m = Monomial(fields)
		m.simplify()
		p = Polynomial(m)
		
		if i.degree() == 2:

			f1 = i.fields[0]
			f2 = i.fields[1]

			#sigma = Field('\\sigma', pos, dimension=6)
			#sigma = -taulx + taunx
			if pos == 'x':
				sigma = -taulx + taunx
				sigma_phi = (phibx*lphix + lphibx*phix)*(-half) + nphibx*nphix
				sigma_psi = (psibx*lpsix + lpsibx*psix)*(-half) + npsibx*npsix
			else:
				raise 'Cannot localize to '+pos

			lf1 = Field('-\\Delta ' + f1.name, pos, fermion=f1.fermion, dimension=f1.dimension+2)
			lf2 = Field('-\\Delta ' + f2.name, pos, fermion=f2.fermion, dimension=f2.dimension+2)

			# (1.49) in rg-loc
			if f1.pos == f2.pos:
				if f1.pos != pos: 
					dxy = Coupling('(' + pos + '_1-' + f1.pos +'_1)^2')
					if f1.fermion:
						p += sigma_psi * dxy
					else:
						p += sigma_phi * dxy

					# the factor 1/2 is because we are doing this for both psi and phi!
					# p += sigma * dxy * half
			# (1.52) in rg-loc
			elif f1.dimension+f2.dimension+2 <= 4:
				if f1.pos < f2.pos:
					dxy = Coupling('(' + f1.pos + '_1-' + f2.pos +'_1)^2')
				else:
					dxy = Coupling('(' + f2.pos + '_1-' + f1.pos +'_1)^2')
				if f1.pos == pos:
					# p += f1 * lf2 * dxy * half
                                        p += - f1 * lf2 * dxy * half  # correction to line above
				else:
					# p += lf1 * f2 * dxy * half
                                        p += - lf1 * f2 * dxy * half  # correction to line above

                        # Loc for (nabla phi) phi
                        # as in greek.tex Section~\ref{sec:nablaphiphi}.
                        # adds correction to p for moving field not at x to x
                        
                        elif f1.dimension+f2.dimension == 3:
                                if i.fields[0].pos == i.fields[1].pos:
                                        return p
                                # case f1.pos = pos; handle other case by switching
                                if f1.pos == pos:
                       			fnew1 = i.fields[0]
			                fnew2 = i.fields[1]
                                else:
                                        fnew1 = i.fields[1]
			                fnew2 = i.fields[0]
                                dxy = '(\\nabla_x^t)^{-1}'
                                # nabla phi_x \phi_y -> (nabla_x^t)^{-1}(nabla phi)_x^2
			 	if fnew2.dimension == 1:
                                        nfnew2pos = Field('\\nabla ' + fnew2.name, pos, fermion=fnew2.fermion, dimension=fnew2.dimension+1)
			 		p += fnew1*nfnew2pos*dxy
                                # nabla phi_y phi_x  -> - (nabla_x^t)^{-1} (-Delta phi_x) phi_x
			 	else:
                                        lfnew2pos = Field(fnew2.name.replace('\\nabla ','-\\Delta '), pos, fermion=fnew2.fermion, dimension=fnew2.dimension)
			 	        p += - fnew1*lfnew2pos*dxy

		return p

	# i is a polynomial: localize each monomial separately
	elif isinstance(i, Polynomial):

		p = Polynomial()
		for g, m in i.monoms:
			p += Loc(m, pos, dim=dim)*g
		return p

	else:
		raise TypeError("%s must be Polynomial or Monomial." % i)


# Finally, calculate "P" as in rg-pt.
Vx  = tau2x * g + taux * nu     + taulx * z + taunx * y
Vy  = tau2y * g + tauy * nu     + tauly * z + tauny * y
EVx = tau2x * g + taux * nuplus + taulx * z + taunx * y
EVy = tau2y * g + tauy * nuplus + tauly * z + tauny * y

# 1st order
# Px = Vx.E(C)
# 2nd order
Px = (Loc(F(Wplus, EVx, EVy),'x') - Loc(F(W, Vx, Vy),'x').E(C))*half

# change to "tau"-basis
taux_n  = Field('\\tau', 'x')
taulx_n = Field('\\tau_\\Delta', 'x')
taunx_n = Field('\\tau_{\\nabla\\nabla}', 'x')
tau_coupling  = Px.proj(phibx*phix)
tau2_coupling = Px.proj(phibx*phix*phibx*phix)
taul_coupling = Px.proj(phibx*lphix) * 2
taun_coupling = Px.proj(nphibx*nphix)
Px = Px - taux  * tau_coupling  + taux_n    * tau_coupling
Px = Px - tau2x * tau2_coupling + taux_n**2 * tau2_coupling
Px = Px - taulx * taul_coupling + taulx_n   * taul_coupling
Px = Px - taunx * taun_coupling + taunx_n   * taun_coupling
Px.simplify()

print Px
