#!/usr/bin/python

# wick.py: Compute moments of Gaussian integrals using Wick's formula
# (requires Python 2.6 or higher)

# Roland Bauerschmidt <brt@math.ubc.ca>
# David C. Brydges <db5d@math.ubc.ca>
# Gordon Slade <slade@math.ubc.ca>

import string
from numbers import Real, Rational
from fractions import Fraction


__FAST__=1


def factorial(n):
	f = 1
	while (n > 0):
		f = f * n
		n = n - 1
	return f

# a very simple permutation class
class Permutation(list):

	def __init__(self, arg):
		if isinstance(arg, int):
			self[:] = range(0,arg)
		else:
			self[:] = arg

	def sign(self):
		sign = 1
		tovisit = range(0, len(self))
		while len(tovisit) > 0:
			cycle = []
			p = tovisit[0]
			while not p in cycle:
				cycle.append(p)
				tovisit.remove(p)
				p = self[p]
			if len(cycle) % 2 == 0:
				sign *= -1
		return sign

# field polynomial like: g phi_x^4+ nu phi_x^2 + + g phi_y^4 + ...
class Polynomial:

	# self.monoms is a list of (coupling, monomial)

	def __init__(self, arg=None):

		self._iszero = 0

		if arg == None:
			self.monoms = []
		elif isinstance(arg, Monomial):
			if arg.iszero():
				self.monoms = []
				self._iszero = 1
			else:
				self.monoms = [(Coupling(), Monomial(arg))]
		elif isinstance(arg, Polynomial):
			self._iszero = arg._iszero
			self.monoms = []
			for g, m in arg.monoms:
				self.monoms.append((Coupling(g), Monomial(m)))
		elif isinstance(arg, list):
			self.monoms = arg
		elif isinstance(arg, Real) or isinstance(arg, Coupling):
			if arg == 0:
				self.monoms = []
				self._iszero = 1
			else:
				self.monoms = [Coupling(arg), Monomial()]
		else:
			raise TypeError('can\'t initialize polynomial with %s' % str(arg))

	def iszero(self):

		self.simplify()
		return self._iszero

	def __add__(self, other):

		if self.iszero():
			return other
		if other.iszero():
			return self

		p = Polynomial()
		for g, m in self.monoms + other.monoms:
			p.monoms.append((Coupling(g), Monomial(m)))
		return p

	def __sub__(self, other):
		minus = Coupling(-1)
		return self + other * minus

	def __mul__(self, other):

		if self.iszero():
			return self

		if isinstance(other, Polynomial):
			if other.iszero():
				return Polynomial(0)

			monoms = []
			for g1, f1 in self.monoms:
				for g2, f2 in other.monoms:
					f1f2 = (f1*f2).simplify()
					if not f1f2.iszero():
						monoms.append((g1*g2, f1*f2))
			return Polynomial(monoms).simplify()

		if isinstance(other, Real) or isinstance(other, str):
			return self*Coupling(other)

		if isinstance(other, Coupling):
			if other.iszero():
				return Polynomial(0)

			monoms = []
			for g, m in self.monoms:
				monoms.append((g*other, m))
			return Polynomial(monoms).simplify()

		raise TypeError('can\'t multiply polynomial by %s' % str(other))

	def __neg__(self):

		return self * (-1)

	def simplify(self):

		# self is 0 or 1
		if len(self.monoms) == 0:
			return self

		e = {}
		d = {}
		s = {}

		for g, m in self.monoms:

			msign, mstr = m.get_sign_str()
			if msign == 0:
				continue

			if e.has_key(mstr):
				e[mstr] += g*msign*s[mstr]
			else:
				s[mstr] = msign
				d[mstr] = m
				e[mstr] = g

		l = e.items()
		l.sort()

		self.monoms = []
		for f, g in l:
			if not g.iszero():
				self.monoms.append((g, d[f]))
		if len(self.monoms) == 0:
			self._iszero = 1

		return self

	def __repr__(self):
		r = ''
		e = {}

		self.simplify()
		if self.iszero():
			return '0'
		if len(self.monoms) == 0:
			return '1'

		deg = 0
		nprocessed = 0
		while nprocessed < len(self.monoms):

			for g, m in self.monoms:

				if m.degree() != deg:
					continue
				nprocessed += 1
				
				if g.iszero():
					continue

				msign, mstr = m.get_sign_str()
				#g.simplify()
				gsigned = g*msign
	
				if r != '':
					r += '\n\n'
					r += ' + '
				r += str(gsigned) + ' ' + mstr
				

			deg += 1

		return r

	def proj(self, monomial):

		r = None
		for g, f in self.monoms:
			if str(f) == str(monomial):
				if r == None:
					r = g
				else:
					r += g
		if r == None:
			r = Coupling(0)
		return r

	def Laplace(self, C):

		p = Polynomial(0)
		for g, f in self.monoms:
			p += f.Laplace(C) * g
		return p

	def E(self, C, sign=1):

		p = Polynomial()
		for g, f in self.monoms:
			e = f.E(C, sign=sign)
			p += e*g
		return p.simplify()

	def wick(self, C):

		return self.E(C, -1)

	def Loc(self, pos):

		p = Polynomial()
		for g, m in self.monoms:
			p += m.Loc(pos)*g
		return p

	def reduceorder(self, order):

		monoms = []
		for g, f in self.monoms:
			r = g.reduceorder(order)
			if not r.iszero():
				monoms.append((r, f))
		self.monoms = monoms
		return self

# field monomial like phi_x^2 phi_y^2 (without prefactor)
class Monomial(Polynomial):

	def __init__(self, arg=None):
		self._iszero = 0
		
		if arg == None:
			self.fields = []
		elif isinstance(arg, list):
			self.fields = arg
		elif isinstance(arg, Monomial):
			self.fields = list(arg.fields)
			self._iszero = arg._iszero
		elif arg == 0:
			self.fields = []
			self._iszero = 1
		else:
			raise TypeError('can\'t initialize monomial with %s' % str(arg))

		# a monomial can also be considered a polynomial with coupling '' = 1
		self.monoms = [(Coupling(1), self)]

	def dimension(self):
		d = 0
		for f in self.fields:
			d += f.dimension
		return d

	def degree(self):
		#if len(self.fields) == 1 and self.fields[0] == '':
		#	return 0
		#if len(self.fields) == 0:
		#	return -1

		self.simplify()
		if self._iszero == 1:
			return -1
		return len(self.fields)

	def iszero(self):
		self.simplify()
		return self._iszero
		#return len(self.fields) == 0
		#return (self.degree() < 0)

	def simplify(self):

		# Fermions anticommute; so if any Fermion appears more than
		# once, the entire monomial vanishes.
		fermions = []
		for f in self.fields:
			if f.fermion:
				if str(f) in fermions:
					# monomial becomes zero
					self.fields = []
					self._iszero = 1
					return self
				fermions.append(str(f))
		return self

	def __repr__(self):
		sign, r = self.get_sign_str()
	
		if sign == 1:
			return r
		elif sign == -1:
			return '-'+r
		else:
			return '0'

	# get simplified string expression for monomial
	#
	# for example:
	#   [phi_x, phi_y, phi_x] --> phi_x^2 phi_y
	#   [psi_x, psi_y, psi_x] --> - psi_x^2 psi_y
	#
	def get_sign_str(self):

		self.simplify()

		#if self._iszero:
		#	print 'XXX'

		if self._iszero:
			return (0, '')

		# the overall sign may be changed by ordering fermions
		r = ''

		powers = {}
		fposs = {}
		fpos = 0 # keep track of number of fermions processed
		for f in self.fields:
			if powers.has_key(str(f)):
				powers[str(f)] += 1

				# if a fermion occurs more than once, the expression is 0
				if f.fermion:
					return (0, '')
			else:
				powers[str(f)] = 1
				# record whether fermion is at even or odd position
				if f.fermion:
					fposs[str(f)] = fpos
				else:
					fposs[str(f)] = -1
			
			if f.fermion:
				fpos += 1

		keys = powers.keys()
		keys.sort()
		fperm = Permutation(fpos) # permution: new fermion position -> old fermion position
		fpos = 0
		for k in keys:
			
			if r != '': r += ' '

			if powers[k] == 1:
				r += k
			else:
				r += k+'^'+str(powers[k])
		
			# k is a fermion; increase fermion position by 1
			if fposs[k] != -1:
				fperm[fpos] = fposs[k]
				fpos += 1

		return (fperm.sign(), r)

	def __mul__(self, other):
		if isinstance(other, Monomial) or isinstance(other, Field):
			return Monomial(self.fields+other.fields)
		elif isinstance(other, Coupling):
			return Polynomial([(other, self)])
		elif isinstance(other, Real) or isinstance(other, str):
			return self*Coupling(other)
		raise TypeError('can\'t multiply monomial by %s' % str(other))

	def Laplace(self, C):

		# the following assumes that the monomial is simplified
		# (otherwise it gives a wrong answer)
		self.simplify()

		if len(self.fields) < 2:
			return Polynomial(0)

		monoms = []
		fn = 0
		for n in range(0,len(self.fields)):
			fm = 0
			for m in range(0,n):

				if C(self.fields[n], self.fields[m]):
					f1 = self.fields[n]
					f1pos = fn
					f2 = self.fields[m]
					f2pos = fm
				elif C(self.fields[m], self.fields[n]):
					f2 = self.fields[n]
					f2pos = fn
					f1 = self.fields[m]
					f1pos = fm
				else:
					if self.fields[m].fermion:
						fm += 1
					continue

				c = C(f1, f2)
				fields = list(self.fields)
				fields.pop(n)
				fields.pop(m)

				# for fermions: compute the sign to permute the two
				# fields that are differentiated to the front
				if f1.fermion:
					if f1pos < f2pos:
						s = f1pos + f2pos - 1
					else:
						s = f1pos + f2pos - 2
					if s % 2 == 0:
						sign = 1
					else:
						sign = -1
				else:
					sign = 1

				monoms.append((c*sign, Monomial(fields)))

				if self.fields[m].fermion:
					fm += 1
			if self.fields[n].fermion:
				fn += 1
		if len(monoms) == 0:
			return Polynomial(0)
		return Polynomial(monoms)
				
	def E(self, C, sign=1):
		p = Polynomial([(Coupling(1), Monomial(self))])
		for n in range(1,len(self.fields)/2+1):
			l = Monomial(self)
			for m in range(0,n):
				l = Polynomial(l.Laplace(C))
				l *= sign
			if n > 1:
				l *= Fraction(1, factorial(n))
			p += l
		return p

	def wick(self, C):
		return self.E(C, -1)

	def Loc(self, pos, dim=4):
		if self.dimension() > dim:
			return Monomial(0)

		fields = []
		for f in self.fields:
			fields.append(f.Loc(pos))
		m = Monomial(fields)
		m.simplify()
		return m

# a single field like phi_x
class Field(Monomial):
	
	def __init__(self, name, pos=None, fermion=0, dimension=1):
		#if type(names) == tuple:
		#	name = names[0]
		#	print type(name)
		#else:
		#	name = names
		if pos == None and isinstance(name, Field):
			self.name = name.name
			self.pos = name.pos
			self.fermion = name.fermion
			self.dimension = name.dimension
		elif isinstance(name, str) and pos != None:
			self.name = name
			self.pos = pos
			self.fermion = fermion
			self.dimension = dimension
		else:
			raise TypeError('can\'t initialize field with %s' % str(name))

		# field is also a monomial (and a polynomial)
		self.fields = [self]
		self.monoms = [(Coupling(1), Monomial(self.fields))]
		self._iszero = 0
	
	def __repr__(self):
		return self.name+'_'+self.pos

	# returns a monomial
	def __pow__(self, p):
		m = []
		for i in range(0,p):
			m.append(self)
		return Monomial(m)

	def Loc(self, pos):
		return Field(self.name, pos=pos, fermion=self.fermion, dimension=self.dimension)

# a coupling constant; can be a product and sum of several terms
class Coupling:

	# self.couplings is a list (=sum) of tuples ([c1, c2, ...], multiplicity)

	def __init__(self, arg=None, arg2=1, priority=0):
		self._iszero = 0

		if arg == None:
			self.couplings = [([], 1)]
		elif isinstance(arg, Coupling):
			self.couplings = arg.couplings
		elif isinstance(arg, list):
			self.couplings = arg
		elif isinstance(arg, str):
			self.couplings = [([arg], arg2)]
		elif isinstance(arg, Real):
			if arg == 0:
				self.couplings = []
				self._iszero = 1
			else:
				self.couplings = [([], arg)]
		else:
			raise TypeError('can\'t initialize field with %s' % str(arg))

	# create power representation of list of coupling constants
	# e.g.: l=['g','g','h'] ----> 'g^2 h'
	def __reprpower(self, l):

		# this brought down the running time from 20 seconds to 15 seconds
		# but it is negligible to the improvement in reprmult
		# if __FAST__:
		#	l.sort()
		#	return string.join(l, ' ')

		r = '' # representation of the empty list
		dp = {} # directory of powers
		for g in l:
			if dp.has_key(g):
				dp[g] += 1
			else:
				dp[g] = 1

		# list of powers: e.g. {'g': 2, 'h': 1}
		lp = dp.items()
		lp.sort()
		lp.reverse() # so lower case coupling constants are first

		# create r='g^2 h' from lp={'g': 2, 'h': 1}
		for g, n in lp:
			if g == '':
				continue
			if r != '':
				r += ' '
			if n==1:
				r += g
			elif g.find(' ') >= 0:
				r += '['+g+']^'+str(n)
			else:
				r += g+'^'+str(n)

		return r

	# collect multiplicities: e.g.
	#   l = [([g,g,g], 3), ([g,g,g], 2)]
	#   ---> d = {'g^3': 5}
	def reprmult(self):
		dm = {} # directory of multiplicities
		if __FAST__:
			do = {} # store g^3 = [g,g,g]

		for gl, x in self.couplings:
			gl.sort()

			g = self.__reprpower(gl)
			if dm.has_key(g):
				dm[g] += x
			else:
				dm[g] = x
				if __FAST__:
					do[g] = gl

		# whenever we have collected the multiplicites, it seems like a good idea
		# to simplify the internal data using it
		# this improves running time from 20 seconds to 0.3 seconds
		if __FAST__:
			self.couplings = [(do[k], dm[k]) for k in dm.keys()]

		return dm

	def __repr__(self):

		# directory of multiplicities: e.g. dm = {'g^3': 5}
		dm = self.reprmult()

		# connect dm with appropriate + and -: e.g.
		#   d = {'g^3': 5, 'h': 1}
		#   ---> s = '5 g^3 + h'

		s = '' # result
		count = 0 # number of summands in s
		lm = dm.items() # list of multiplicities
		lm.sort()
		for g, x in lm:
			if x == 1:
				if s != '': s += ' + '
				s += g+ '\n'
				count += 1
			elif x == -1:
				s += ' - '+g +'\n'
				count += 1
			elif x != 0:
				if isinstance(x, Rational) and x.denominator != 1:
					a = '\\frac{%d}{%d}' % (abs(x.numerator), x.denominator)
				else:
					a = str(abs(x))
				if x > 0:
					s += ' + '+a+' '+g+'\n'
				else:
					s += ' - '+a+' '+g+'\n'
				count += 1

		if count == 0:
			return '0'
		elif count > 1:
			s = '[' + s + ']'
		return s

	def iszero(self):

		dm = self.reprmult()
		for g, x in dm.items():
			if x != 0:
				return 0
		return 1

	def __add__(self, other):

		#if self.iszero():
		#	return other
		#if other.iszero():
		#	return self

		if isinstance(other, str):
			if other == '':
				return self
			else:
				return Coupling(self.couplings + [([other], 1)])
		elif isinstance(other, Coupling):
			return Coupling(self.couplings + other.couplings)

	def __sub__(self, other):
		return self + (other * Coupling(-1))

	def __mul__(self, other):

		if self.iszero():
			return self

		couplings = []

		if isinstance(other, Real):
			if other == 0:
				return self
			for gl, x in self.couplings:
				couplings.append((gl, x*other))
		elif isinstance(other, Coupling):
			if other.iszero():
				return other
			for gl, x in self.couplings:
				for hl, y in other.couplings:
					couplings.append((gl+hl, x*y))
		else:
			raise TypeError('can\'t multiply coupling by %s' % str(other))

		return Coupling(couplings)

	def reduceorder(self, order):
		couplings = []
		for c,e in self.couplings:
			o = 0
			for s in c:
				if s[0].islower() or s[0] == '{':
					o += 1
			if o <= order:
				couplings.append((c,e))
		self.couplings = couplings
		return self
